local moonbridge_io = require "moonbridge_io"
local extos = require 'extos'

local sandbox = require "sandbox"

os.execute("mkdir -p state")

local DEAD = {}
local SLEEP_UNTIL = {}

local container_module = {}

local loaded_containers = {}
local running_coroutines = {}
local new_coroutines = {}

local inputs = {}
local outputs = {}
local wait_until = math.huge
local news = false

function resume_coroutine(coro)
  local r0, r1, r2, r3, r4 = coroutine.resume(coro)
  if not r0 then
    error("error in coroutine: " .. debug.traceback(coro, r1))
  end
  if r1 == DEAD then
    running_coroutines[coro] = nil
    new_coroutines[coro] = nil
    news = true
  elseif r1 == SLEEP_UNTIL then
    wait_until = math.min(wait_until, r2)
    if r3 then
      news = true
    end
  elseif r4 == moonbridge_io or r4 == "moonbridge_io" then
    if not news then
      if r3 then
        news = true
      else 
        if r2 == "r" then
          inputs[r1] = true
        elseif r2 == "w" then
          outputs[r1] = true
        else
          error("illegal socket operation mode, bad bad!")
        end
      end
    end
  else
    error("illegal yield, bad bad!")
  end
end

local function sleep_until(timestamp)
  local first = true
  while timestamp > moonbridge_io.timeref() do
    coroutine.yield(SLEEP_UNTIL, timestamp, first)
    first = false
  end
end

local function spawn(func)
  local coro = coroutine.create(function()
    func()
    return DEAD
  end)
  new_coroutines[coro] = true
  return coro
end

function container_module.call_container_func(container_id, func)
  local container = loaded_containers[container_id]
  local old_caller = container.old_caller
  container.caller = nil
  local result = container_module.call(container_id, func, args)
  container.caller = old_caller
  return result
end


local function load_module(container_id, module_id)
  io.stderr:write("[Container] load container ", container_id, " with module ", module_id, "\n")
  local fh = io.open("module/" .. module_id .. ".lua", "r")
  if not fh then
    error("[MAIN] ERROR opening module file: " .. module_id)
  end

  local module = fh:read("*a")
  fh:close()
  if not module then
    error("[MAIN] ERROR reading module file: " .. module_id)
  end

  local container = {
    id = container_id,
    module_id = module_id,
    modules = {}
  }

  local sandbox_G = sandbox.new()
  setmetatable(sandbox_G, {
    __index = function(self, key)
      if key == "calling_module" then
        return loaded_containers[container_id].caller
      end
    end
  })
  sandbox_G.module = setmetatable({}, {
    __index = function(self, interface)
      local tmp = container.modules[interface]
      if not tmp then
        -- TODO evtl. verbessern
        io.stderr:write("[Container] container " .. container_id .. " has no interface " .. interface, "\n")
        return
      end
      local r
      if type(tmp) == "table" then
        r = {}
        for module_id in pairs(tmp) do
          r[module_id] = setmetatable({}, {
            __index = function(self, func)
              return function(args)
                --print("CALL", container.id, interface, module_id, func)
                local other_container = loaded_containers[module_id]
                local old_caller = other_container.old_caller
                other_container.caller = container_id
                local result = container_module.call_interface(module_id, interface, func, args)
                other_container.caller = old_caller
                return result
              end
            end
          })
        end
      else
        r = setmetatable({}, {
          __index = function(self, func)
            return function(args)
              --print("CALL1", container.id, interface, func)
              local other_container = loaded_containers[tmp]
              local old_caller = other_container.old_caller
              other_container.caller = container_id
              local result = container_module.call_interface(tmp, interface, func, args)
              other_container.caller = old_caller
              return result
            end
          end
        })
      end
      return r
    end
  })
  sandbox_G.sleep_until = sleep_until
  sandbox_G.spawn = spawn
  -- TODO sicherheit
  sandbox_G.extos = extos
  -- TODO sicherheit
  sandbox_G.io_open = io.open
  if container_id == "state" then
    sandbox_G.load_store = function()
      local fh = io.open("state/store", "r")
      if not fh then
        return nil
      end
      local data = fh:read("*a")
      fh:close()
      return data
    end
    sandbox_G.save_store = function(data)
      local fh = io.open("state/store", "w")
      fh:write(data)
      fh:close()
    end
    sandbox_G.load_state = function(index)
      if not index then
        local fh = io.open("state/current_index", "r")
        if not fh then
          return
        end
        index = fh:read("*a")
        fh:close()
      end
      local fh = io.open("state/" .. index, "r")
      if not fh then
        return nil
      end
      local data = fh:read("*a")
      fh:close()
      return data
    end
    sandbox_G.save_state = function(index, data)
      local fh = io.open("state/" .. index, "w")
      fh:write(data)
      fh:close()
      local fh = io.open("state/current_index", "w")
      fh:write(index)
      fh:close()
    end
    sandbox_G.revert_state = function(index)
      -- TODO delete higher states
      local fh = io.open("state/current_index", "w")
      fh:write(index)
      fh:close()
    end
  end
  local sandbox_func, err_message = load(
    module, 
    "=module/" .. module_id .. ".lua", 
    "t", 
    sandbox_G
  )
  if not sandbox_func then
    error("[MAIN] ERROR loading module " .. module_id .. ": " .. err_message )
  end
  
  local stacktrace
  local success, result, message = xpcall(
    sandbox_func, 
    function(err) 
      stacktrace = debug.traceback() 
      return err
    end
  )
  if not success then
    error("[MAIN] ERROR executing module " .. module_id .. ": " .. result)
  end
  
  container.module = result

  return container
end

local function init_container(container)
  local coro = spawn(function()
    if container.module and container.module.init then
      container.module.init()
    end
  end)
  loaded_containers[container.id] = container
  resume_coroutine(coro)
end

local container_modules = {}

function container_module.load(module_ids)
  loaded_containers = {}
  local containers = {}
  for i, module_id in ipairs(module_ids) do
    local container = load_module(module_id, module_id)
    table.insert(containers, container)
  end
  
  print("[Container] Collecting provided interfaces...")
  
  local provides = {}
  local provides_single = {}
  for i, container in pairs(containers) do
    if container.module.provide then
      for interface_id in pairs(container.module.provide) do
        print("[Container] " .. container.module_id .. " provides " .. interface_id)
        provides[interface_id] = provides[interface_id] or {}
        provides[interface_id][container.module_id] = true
      end
    end
    if container.module.provide_single then
      for interface_id in pairs(container.module.provide_single) do
        print("[Container] " .. container.module_id .. " provides single " .. interface_id)
        provides_single[interface_id] = provides_single[interface_id] or {}
        provides_single[interface_id][container.module_id] = true
      end
    end
  end

  print("[Container] Checking needed interfaces...")

  for i, container in pairs(containers) do
    if container.module.need_single then
      local matching_module = false
      for i, interface_id in ipairs(container.module.need_single) do
        container.modules[interface_id] = {}
        print("[Container] " .. container.module_id .. " need single " .. interface_id)
        local err
        local matching_module
        if provides_single[interface_id] then
          for module, available in pairs(provides_single[interface_id]) do
            if matching_module then
              err = "need_single_but_found_multiple"
            elseif not available then
              err = "need_single_but_already_used"
            else
              container.modules[interface_id][module] = module
              matching_module = module
            end
          end
        end
        if matching_module then
          provides_single[interface_id][matching_module] = false
        end
        if provides[interface_id] then
          local module_count = 0
          local module
          for m in pairs(provides[interface_id]) do
            module_count = module_count + 1
            module = m
          end
          if module_count > 1 then
            err = "need_single_but_multiple_found1"
          elseif module_count == 1 and matching_module then
            err = "need_single_but_multiple_found2"
          elseif module_count == 1 then
            container.modules[interface_id] = module
            matching_module = module
          end
        end
        if not matching_module then
          if not err then
            err = "need_single_but_non_found"
          end
          -- TODO error handling
          print("[Container] ERROR", err)
          os.exit()
        end
      end
    end
  end

  for i, container in pairs(containers) do
    if container.module.uses then
      for i, interface_id in ipairs(container.module.uses) do
        local err
        container.modules[interface_id] = {}
        print("[Container] " .. container.module_id .. " uses " .. interface_id)
        if provides_single[interface_id] then
          for module, available in pairs(provides_single[interface_id]) do
            if not available then
              err = "uses_but_already_used"
            else
              container.modules[interface_id][module] = module
              provides_single[interface_id][module] = false
            end
          end
        end
        if provides[interface_id] then
          for module in pairs(provides[interface_id]) do
            container.modules[interface_id][module] = module
          end
        end
      end
    end
  end

  for i, container in ipairs(containers) do
    init_container(container)
  end
end

function container_module.call_interface(container_id, interface, func, args)
  local container = loaded_containers[container_id]
  if container and container.module and container.module.provide and container.module.provide[interface] and container.module.provide[interface][func] then
    local result = container.module.provide[interface][func](args)
    return result
  elseif container and container.module and container.module.provide_single and container.module.provide_single[interface] and container.module.provide_single[interface][func] then
    local result = container.module.provide_single[interface][func](args)
    return result
  end
end

function container_module.call(container_id, func, args)
  local container = loaded_containers[container_id]
  if container and container.module and container.module[func] then
    local result = container.module[func](args)
    return result
  end
end

local function shutdown()
  for container_id in pairs(loaded_containers) do
    container_module.call(container_id, "on_exit")
  end
end
  
function container_module.next_loop()
  inputs = {}
  outputs = {}
  wait_until = math.huge
  news = false
  for coro in pairs(new_coroutines) do
    running_coroutines[coro] = true
    new_coroutines[coro] = nil
  end
  for coro in pairs(running_coroutines) do
    resume_coroutine(coro)
  end
  if not news and not next(new_coroutines) then
    local wait_time
    if wait_until ~= math.huge then
      wait_time = wait_until - moonbridge_io.timeref()
    end
    local status, err_message, term = moonbridge_io.poll(inputs, outputs, wait_time, true)

    if term then
      shutdown()
    end
    return term
  end
end

return container_module
