local BLOCK_INTERVAL = 60

local state

local blockchain_module = {
  ident = "blockchain",
  uses = {
    "app"
  },
  need_single = {
    "hash",
    "state",
    "manage_state",
    "user"
  },
  provide = {
    api = {},
    blockchain = {},
    mainloop = {}
  }
}

-- Serves a static API file
function blockchain_module.provide.api.file(args)
  if files[args.path] then
    return json.object{
      content_type = "text/html",
      data = files[path].content
    }
  else
    return json.object{
      status = 404
    }
  end
end

-- Serves the blockchain a json document
function blockchain_module.provide.api.get_blocks(args)
  return json.object{
    content_type = "application/json",
    data = state.blocks
  }
end



local blockchain = {}

-- ------------------------------------------------------------------------
-- FRONTEND
-- ------------------------------------------------------------------------

local files = {
  [""] = {
    content = [[
      <h2>Module: Blockchain</h2>
      <pre><ul id="blocks"></ul></pre>
      <script>
        api("POST", "blockchain", { command: "get_blocks" }, function(result) {
          document.getElementById("blocks").innerHTML = "";
          for (var i = 0; i < result.length; i++) {
            var block = result[i];
            var li_el = document.createElement("li");
            li_el.innerHTML = JSON.stringify(block, null, 2);
            document.getElementById("blocks").appendChild(li_el);
          }
        })
      </script>
    ]]
  }
}


function blockchain_module.provide.api.on_api_get_file(args)
  print("[Blockchain] on api get file '" .. args.file .. "'")
  return files[args.file]
end

function blockchain_module.provide.api.on_api_call(args)
  local method = args.method
  local data = args.data
  local command = data.command
  if command == "get_blocks" then
    return state.blocks
  elseif command == "new_transaction" then
    local transaction = data.transaction
    transaction.person_id = args.person_id
    transaction.signature = module.user.sign_transaction{ transaction = transaction, user_id = args.user_id }
    blockchain.add_transaction(transaction)
    return transaction
  end
end

local transaction_modules = { "person", "demo01" }

-- ------------------------------------------------------------------------
-- ADD / PROCESS TRANSACTIONS
-- ------------------------------------------------------------------------

local function process_transaction(transaction_id, transaction)
  for j, transaction_module in ipairs(transaction_modules) do
    print("[Blockchain] execute process_transaction for module " .. transaction_module)
    module.app[transaction_module].process_transaction{ transaction_id = transaction_id, transaction = transaction }
  end
end

local function process_transactions(transactions)
  for transaction_id, transaction in pairs(transactions) do
    process_transaction(transaction_id, transaction)
  end
end

function blockchain.add_transaction(transaction)
  local transaction_id = module.hash.hash_object{ obj = transaction }
  local last_block = state.blocks[#(state.blocks)]
  local current_timeindex = math.floor(os_time() / BLOCK_INTERVAL) + 1
  if last_block.timeindex < current_timeindex then
    local block = json.object{
      index = (last_block and last_block.index or 0) + 1,
      transactions = json.object(),
      timeindex = current_timeindex,
      interval = BLOCK_INTERVAL,
    }
    state.blocks[#(state.blocks) + 1] = block
    last_block = block
  end
  last_block.transactions[transaction_id] = transaction
  process_transaction(transaction_id, transaction)
  print("[Blockchain/add_transaction] added transaction: ", tostring(last_block.index), " ", json.export(state))
  module.manage_state.set_index{ index = last_block.index }
  return true
end

-- ------------------------------------------------------------------------
-- GET BLOCKS
-- ------------------------------------------------------------------------

function blockchain_module.provide.blockchain.get_blocks(args)
  return json.import(json.export(state.blocks))
end

-- ------------------------------------------------------------------------
-- REPLACE BLOCKS
-- ------------------------------------------------------------------------

function blockchain_module.provide.blockchain.replace_blocks(args)
  print("[Blockchain/replace] replace blocks called")
  if not args.blocks then
    return
  end
  local start_index = math.huge
  local need_replacement = false
  for i, block in ipairs(args.blocks) do
    local old_block = state.blocks[block.index]
    local old_block_hash = module.hash.hash_object{ obj = old_block }
    local block_hash = module.hash.hash_object{ obj = block }
    if not old_block or old_block_hash ~= block_hash then
      need_replacement = true
      start_index = math.min(start_index, block.index)
    end
  end
  if need_replacement then
    print("[Blockchain/replace] rollback to state " .. start_index - 1)
    state = module.manage_state.revert{ index = start_index - 1 }
    if not state.blocks then
      state.blocks = json.array()
    end
    for i = start_index, #(args.blocks) do
      local block = args.blocks[i]
      print("[Blockchain/replace] block " .. block.index .. " need replacement")
      state.blocks[block.index] = block
      -- rollback state
      module.manage_state.set_index{ index = block.index }
      -- apply new state
      process_transactions(block.transactions)
      print("[Blockchain/replace] block " .. block.index .. " replaced")
    end
    for i = #(args.blocks) + 1, #(state.blocks) do
      state.blocks[i] = nil
    end
  end
end

-- ------------------------------------------------------------------------
-- FIRST BLOCK
-- ------------------------------------------------------------------------

local function first_block()
  print("[Blockchain/first_block] started")
  module.state.set_index{ index = 0 }
  state.blocks[1] = json.object{
    index = 1,
    transactions = json.object(),
    timeindex = math.floor(os_time() / BLOCK_INTERVAL) + 1,
    interval = BLOCK_INTERVAL,
    previous_hash = "-",
  }
end

-- ------------------------------------------------------------------------
-- INIT
-- ------------------------------------------------------------------------

function blockchain_module.init()
  spawn(function()
    print("[Blockchain/init] main")
    state = module.state.get_state()
    if not state.blocks then
      state.blocks = json.array()
      first_block()
    end
    --spawn(new_block_loop)
  end)
end

return blockchain_module
