local BLOCK_INTERVAL = 60

local consensus_module = {
  ident = "consensus",
  need_single = {
    "blockchain",
    "choose_node",
    "hash"
  },
  provide_single = {
    sync = {}
  }
}

consensus_module.module_name = "Consensus"


-- ------------------------------------------------------------------------
-- MERGE BLOCKCHAINS
-- ------------------------------------------------------------------------

local function merge_blockchains(local_blockchain, partial_blockchains)
  print("[Consensus/merge_blockchains] started")
  
  local weight_sum = 0
  
  local transactions = {}
  local timeindexes = {}
  
  for dummy, partial_blockchain in pairs(partial_blockchains) do
    weight_sum = weight_sum + partial_blockchain.weight
    for dummy, block in ipairs(partial_blockchain.blocks) do
      local timeindex = tostring(block.timeindex)
      for transaction_id, transaction in pairs(block.transactions) do
        transactions[transaction_id] = transaction
        timeindexes[timeindex] = timeindexes[timeindex] or {}
        timeindexes[timeindex][transaction_id] = (timeindexes[timeindex][transaction_id] or 0) + partial_blockchain.weight
      end
    end
  end
  
  for dummy, block in ipairs(local_blockchain) do
    local timeindex = tostring(block.timeindex)
    for transaction_id, transaction in pairs(block.transactions) do
      transactions[transaction_id] = transaction
      timeindexes[timeindex] = timeindexes[timeindex] or {}
      local weight = weight_sum
      for dummy, weight in pairs(timeindexes[timeindex]) do
        weight = weight - 1
      end
      timeindexes[timeindex][transaction_id] = (timeindexes[timeindex][transaction_id] or 0) + weight
      if not (timeindexes[timeindex][transaction_id] > weight_sum / 2) then
        timeindexes[timeindex][transaction_id] = timeindexes[timeindex][transaction_id] + 1
      end
    end
  end
  
  local timeindexes_list = {}
  for timeindex in pairs(timeindexes) do
    table.insert(timeindexes_list, tonumber(timeindex))
  end
  table.sort(timeindexes_list)
  
  local new_blockchain = json.array()
  local transaction_without_block = json.object()

  for i, timeindex in ipairs(timeindexes_list) do
    local block = json.object{ 
      index = #new_blockchain + 1,
      timeindex = timeindex,
      interval = BLOCK_INTERVAL,
      transactions = json.object(),
      previous_hash = #new_blockchain > 0 and module.hash.hash_object{ obj = new_blockchain[#new_blockchain] } or "-"
    }
    local transaction_found
    for transaction_id, weight in pairs(timeindexes[tostring(timeindex)]) do
      local transaction = transactions[transaction_id]
      if weight > weight_sum / 2 or i == #timeindexes_list then
        -- transaction accepted for this timeindex
        block.transactions[transaction_id] = transaction
        transaction_without_block[transaction] = nil
        transaction_found = true
      else
        -- transaction not accepted for this timeindex
        transaction_without_block[transaction] = true
      end
    end
    if transaction_found then
      new_blockchain[#new_blockchain + 1] = block
    end
  end
  
  local block = json.object{
    index = #new_blockchain + 1,
    timeindex = math.floor(os_time() / BLOCK_INTERVAL) + 1,
    interval = BLOCK_INTERVAL,
    transactions = json.object(),
    previous_hash = #new_blockchain > 0 and module.hash.hash_object{ obj = new_blockchain[#new_blockchain] } or "-"
    }
  local transaction_without_block_found
  for transaction in pairs(transaction_without_block) do
    block.transactions[transaction.id] = transaction
    transaction_without_block_found = true
  end
  if transaction_without_block_found then
    new_blockchain[#new_blockchain + 1] = block
  end
  
  return new_blockchain
  
end

-- ------------------------------------------------------------------------
-- P2P SYNC
-- ------------------------------------------------------------------------

function consensus_module.has_sync()
  return true
end

local sync_in_blocks = json.object()

function consensus_module.provide_single.sync.sync_in(args)
  print("[Consensus/sync_in]")
  local node_id = args.node_id
  local iteration = args.iteration
  local request = args.request
  if iteration == 1 then
    sync_in_blocks[node_id] = module.blockchain.get_blocks()
  end
  local blocks = sync_in_blocks[node_id]
  local count = 2 ^ (iteration - 1)
  local stop = #blocks - count + 1
  local start = math.min(stop - count + 1, 1)
  local result = json.object()
  result.blocks = json.array()
  for i = start, stop do
    table.insert(result.blocks, blocks[i])
  end
  return result
end

local sync_blockchains = json.object()

function consensus_module.provide_single.sync.sync_out(args)
  print("[Consensus/sync_out]")
  local local_blocks = module.blockchain.get_blocks()
  local remote_blocks = args.result and args.result.blocks or json.array()
  for i, block in ipairs(sync_blockchains[args.node_id] or {}) do
    table.insert(remote_blocks, block)
  end
  local matched = false
  for i = #remote_blocks, 1, -1 do
    if not matched then
      local remote_block = remote_blocks[i]
      local local_block = local_blocks[remote_block.index]
      if local_block then
        local local_hash = module.hash.hash_object{ obj = local_block }
        local remote_hash = module.hash.hash_object{ obj = remote_block }
        if local_hash == remote_hash then
          matched = i
        end
      end
    end
  end
  if matched then
    local tmp = json.array()
    for i = matched, #remote_blocks do
      table.insert(tmp, remote_blocks[i])
    end
    remote_blocks = tmp
  end
  sync_blockchains[args.node_id] = json.object{
    weight = args.weight,
    blocks = remote_blocks
  }
--  if not matched and (not remote_blocks[1] or remote_blocks[1].index > 1) then
  if args.iteration == 1 or not matched and remote_blocks[1] and remote_blocks[1].index > 1 then
    return { get_blocks = true }
  end
end

function consensus_module.provide_single.sync.sync_out_completed()
  print("[Consensus/sync_completed] merging blockchains")
  local blocks = module.blockchain.get_blocks()
  local merged_blocks = merge_blockchains(blocks, sync_blockchains)
  module.blockchain.replace_blocks{ blocks = merged_blocks }
  sync_blockchains = json.object()
end

function consensus_module.provide_single.sync.sync_out_cancelled()
  print("[Consensus/sync_cancelled] sync cancelled")
  sync_blockchains = json.object()
end

  
return consensus_module
