local node_module = {
  ident = "node",
  need_single = {
    "store",
    "pki",
    "init_sync"
  },
  provide = {
    node = {},
    api = {},
  },
  provide_single = {
    sync = {}
  }
}
local store

-- ------------------------------------------------------------------------
-- FRONTEND
-- ------------------------------------------------------------------------

local files = {
  [""] = {
    content = [[
      <h2>Module: Node</h2>
      <div>Local node ID: <span id="node_id"></span></div>
      <h3>Node list</h3>
      <div id="nodes"></div>
      <script>
        api("POST", "node", { command: "get_local" }, function(result) {
          document.getElementById("node_id").innerHTML = result.id;
        })
        function updateNodes() {
          api("POST", "node", { command: "get_nodes" }, function(result) {
            console.log("nodes", result);
            document.getElementById("nodes").innerHTML = "";
            for (node_id in result.nodes) {
              var config = result.nodes[node_id];
              var li_el = document.createElement("li");
              li_el.innerHTML = node_id + " = " + JSON.stringify(config);
              document.getElementById("nodes").appendChild(li_el);
            }
          })
        }
        updateNodes();
        
      </script>
    ]]
  }
}

function node_module.provide.api.on_api_call(args)
  print("[Node/on_api_call] started")
  local method = args.method
  local data = args.data
  local command = data.command
  if command == "get_local" then
    print("[Node/on_api_call] serving command get_local")
    return json.object{
      id = store.local_node.id
    }
  elseif command == "get_nodes" then
    print("[Node/on_api_call] serving command get_nodes")
    return json.object{
      nodes = store.nodes
    }
  end
end

function node_module.provide.api.on_api_get_file(args)
  print("[Node] on api get file '" .. args.file .. "'")
  return files[args.file]
end

-- ------------------------------------------------------------------------
-- GET LOCAL NODE ID
-- ------------------------------------------------------------------------

function node_module.provide.node.get_local()
  return json.object{
    id = store.local_node.id
  }
end

function node_module.provide.node.get_config(args)
  return store.nodes[args.node_id]
end


-- ------------------------------------------------------------------------
-- SYNC
-- ------------------------------------------------------------------------


function node_module.has_sync()
  return true
end

local function merge_remote_nodes(nodes)
  for node_id, config in pairs(nodes) do
    if not store.nodes[node_id] or config.index > store.nodes[node_id].index then
      local check = module.pki.check_signature{ obj = config, public_key = config.public_key }
      if check.valid == true then
        print(" [Node/merge_remote_nodes] added node: " .. node_id)
        store.nodes[node_id] = config
      else
        print(" [Node/merge_remote_nodes] invalid signature, config dropped")
      end
    end
  end
end

function node_module.provide_single.sync.sync_in(args)
  print("[Node/sync_in]")
  local node_id = args.node_id
  local iteration = args.iteration
  local request = args.request
  local remote_nodes = request.nodes
  merge_remote_nodes(remote_nodes)
  if iteration == 1 then
    return json.object{  nodes = store.nodes }
  end
end

function node_module.provide_single.sync.sync_out(args)
  print("[Node/sync_out]")
  local node_id = args.node_id
  local iteration = args.iteration
  local result = args.result
  if result then
    local remote_nodes = result.nodes
    merge_remote_nodes(remote_nodes)
  end
  if iteration == 1 then
    return json.object{ nodes = store.nodes }
  end
end

function node_module.provide_single.sync.sync_out_completed()
end

function node_module.provide_single.sync.sync_out_cancelled()
end


-- ------------------------------------------------------------------------
-- INIT
-- ------------------------------------------------------------------------

function node_module.init()
  spawn(function()
    print("[Node/init] started")
    print("[Node/init] preparing store")
    store = module.store.get_store()
    if not store.local_node then
      print("[Node/init] local node started for the first time, initializing...")
      local key_pair = module.pki.create_key_pair()
      store.local_node = json.object{
        id = key_pair.fingerprint,
        private_key = key_pair.private_key,
        public_key = key_pair.public_key
      }
      local node_config = json.object{ 
        index = 1,
        public_key = store.local_node.public_key,
        host = P2P_HOST, 
        port = P2P_PORT
      }
      local signature = module.pki.sign{ data = json.export(node_config), private_key = store.local_node.private_key }
      node_config.signature = signature.signature
      local check = module.pki.check_signature{ obj = node_config, public_key = store.local_node.public_key }
      store.nodes = json.object{
        [store.local_node.id] = node_config
      }
      print("[Node/init] local node initialized")
    end
    if INIT_HOST and INIT_PORT then
      module.init_sync.init{ host = INIT_HOST, port = INIT_PORT }
    end
    print("[Node/init] finished")
  end)
end

return node_module
