local SYNC_INTERVAL = 1

local p2p_module = {
  ident = "p2p",
  uses = {
    "sync"
  },
  need_single = {
    "choose_node",
    "node"
  },
  provide = {
    init_sync = {}
  }
}

-- ------------------------------------------------------------------------
-- OUTGOING SYNC
-- ------------------------------------------------------------------------

local next_sync_at = time() + SYNC_INTERVAL

local sync_running = false
local running_sync_nodes = {}

local function sync_out_containers(node_id, iteration, weight, data_in)
  local data_out = json.object()
  local work_to_do = false
  for container_id, container in pairs(module.sync) do
    print("[P2P/sync] process outgoing sync for ", container_id)
    local container_data_in = data_in and data_in[container_id] or nil
    data_out[container_id] = module.sync[container_id].sync_out{ 
      node_id = node_id,
      iteration = iteration,
      weight = weight,
      result = container_data_in
    }
    if data_out[container_id] then
      work_to_do = true
    end
  end
  if not work_to_do then
    data_out = nil
    running_sync_nodes[node_id] = false
    local finished = true
    for node_id, running in pairs(running_sync_nodes) do
      if running then
        finished = false
      end
    end
    if finished then
      for module_id, sync_module in pairs(module.sync) do
        sync_module.sync_out_completed()
      end
    end
  end
  sync_running = false
  return data_out
end

local function sync_out_cancel_containers(node_id, weight, data_in)
  running_sync_nodes[node_id] = nil
  for container_id, container in pairs(module.sync) do
    module.sync[container_id].sync_out_cancelled()
  end
  sync_running = false
end

local function sync_out_node(node_id, host, port, weight)
  spawn(function()
    print("[P2P/sync] new syncer spawned")
    print("[P2P/sync] connecting to " .. host .. ":" .. port)
    local socket = tcpconnect(host, port)
    if socket then
      print("[P2P/sync] connected")
      local i = 1
      local data_out = sync_out_containers(node_id, i, weight)
      while data_out do
        i = i + 1
        local local_node_id = module.node.get_local().id
        local status = socket:flush(json.export(json.object{ data = data_out, node_id = local_node_id }), "\n")
        if not status then
          print("[P2P/sync] connection error")
          sync_out_cancel_containers(node_id, weight)
          return
        end
        local result = socket:read_yield(nil, "\n")
        if result then
          local data_in = json.import(result)
          data_out = sync_out_containers(node_id, i, weight, data_in)
        end
      end
      socket:close()
    else
      print("[P2P/sync] not connected")
    end
  end)
end

local function sync_out()
  print("[P2P/sync] started")
  if sync_running then
    print("[P2P/sync] already running")
    return
  end
  sync_nodes = module.choose_node.get_random_node_ids{ count = 2 }
  for node_id, weight in pairs(sync_nodes) do
    local node_config =  module.node.get_config{ node_id = node_id }
    local host = node_config.host
    local port = node_config.port
    sync_running = true
    sync_out_node(node_id, host, port, weight)
  end
end

local function sync_out_loop()
  print("[P2P/syncloop] started")
  while true do
    print("[P2P/syncloop] sleeping")
    sleep_until(next_sync_at)
    print("[P2P/syncloop] continued")
    if next_sync_at <= time() then
      print("[P2P/syncloop] next sync")
      sync_out()
      next_sync_at = time() + SYNC_INTERVAL
    end
  end
end

-- ------------------------------------------------------------------------
-- INCOMING SYNC
-- ------------------------------------------------------------------------

local function sync_in_containers(node_id, iteration, data_in)
  local data_out = json.object()
  for container_id, container in pairs(module.sync) do
    print("[P2P/sync] process incoming sync for ", container_id)
    local container_data_in = data_in and data_in[container_id] or nil
    if container_data_in then
      data_out[container_id] = module.sync[container_id].sync_in{ node_id = node_id, iteration = iteration, request = container_data_in }
    end
  end
  return data_out
end

local function listen_loop()
  local listen = tcplisten("127.0.0.1", P2P_PORT)
  print("[P2P/listen_loop] listener installed")
  print("[P2P/listen_loop] starting listen loop")
  while true do
    
    local socket
    local first = true
    repeat
      print("[P2P/listen_loop] waiting for new connection")
      coroutine.yield(listen, "r", first, "moonbridge_io")
      print("[P2P/listen_loop] wake up for new connection")
      first = false
      print("[P2P/listen_loop] accepting new connection")
      socket = listen:accept_nb()
      if not socket then
        print("[P2P/listen_loop] no new connection")
      end
    until socket
    
    print("[P2P/listen_loop] accepted new connection")
    spawn(function()
      print("[P2P/listen_loop] new socket connection handler started, reading request from socket")
      local i = 1
      repeat
        local data, status = socket:read_yield(nil, "\n")
        if status ~= "eof" and data then
          local request = json.import(data)
          print("[P2P/listen_loop] request: " .. json.export(request))
          local result = sync_in_containers(request.node_id, i, request.data)
          print("[P2P/listen_loop] sending result: " .. json.export(result))
          local result = json.export(result)
          socket:flush(result, "\n")
        end
        i = i + 1
      until status == "eof"
      print("[P2P/listen_loop] connection finished")
      socket:close()
    end)

  end
end

-- ------------------------------------------------------------------------
-- INIT SYNC
-- ------------------------------------------------------------------------

function p2p_module.provide.init_sync.init(args)
  local host = args.host
  local port = args.port
  sync_running = true
  sync_out_node("", host, port, 1)
end

-- ------------------------------------------------------------------------
-- INIT
-- ------------------------------------------------------------------------

function p2p_module.init()
  spawn(function()
    print("[P2P/init] started")
    print("[P2P/init] starting sync out loop")
    spawn(sync_out_loop)
    print("[P2P/init] starting listen loop")
    listen_loop()
    
  end )
end

return p2p_module
